/*
 * sys_cap.h:
 *
 * See the main source file 'xineliboutput.c' for copyright information and
 * how to reach the author.
 *
 * $Id$
 *
 */

#ifndef _XINELIBOUTPUT_SYS_CAP_H_
#define _XINELIBOUTPUT_SYS_CAP_H_

int have_cap_sys_nice(void);

#endif // _XINELIBOUTPUT_SYS_CAP_H_
