/*
 * vo_osdreorder.h: OSD re-ordering video-out post plugin
 *
 * See the main source file 'xineliboutput.c' for copyright information and
 * how to reach the author.
 *
 * $Id$
 *
 */

#ifndef _XINELIBOUTPUT_VO_OSDREORDER_H
#define _XINELIBOUTPUT_VO_OSDREORDER_H

#include <xine/video_out.h>

vo_driver_t *osd_reorder_init(void);

#endif /* _XINELIBOUTPUT_VO_OSDREORDER_H */
